/*++
Copyright (c) 2018 Microsoft Corporation

Module Name:

    theory_jobscheduler.cpp

Abstract:


Author:

    Nikolaj Bjorner (nbjorner) 2018-09-08.

Revision History:

TODO:

- arithmetic interface

- propagation queue:
  - register theory variables for catching when jobs are bound to resources
  - register bounds on start times to propagate energy constraints
  - more general registration mechanism for arithmetic theory.
- csp_cmds
  - use predicates for add_ feature set? Closed world.
    set up environment in one swoop.
  - interact with opt
- jobs without resources
  - complain or add dummy resource? At which level.

Features:
- properties
- priority
  - try mss based from outside
- job-goal
   - try optimization based on arithmetic solver.
   - earliest start, latest start
- constraint level
- resource groups
  - resource groups like a resource
  - resources bound to resource groups within time intervals
  - job can require to use K resources from a resource group simultaneously.

--*/

#include "ast/ast_pp.h"
#include "smt/theory_jobscheduler.h"
#include "smt/smt_context.h"
#include "smt/smt_arith_value.h"
#include "smt/smt_model_generator.h"

namespace smt {

    theory_var theory_jobscheduler::mk_var(enode * n) {
        theory_var v = theory::mk_var(n);
        return v;
    }

    bool theory_jobscheduler::internalize_term(app * term) {
        context & ctx = get_context();        
        if (ctx.e_internalized(term))
            return true;
        for (expr* arg : *term) {
            if (!ctx.e_internalized(arg)) 
                ctx.internalize(arg, false);
        }        
        enode* e = ctx.mk_enode(term, false, false, true);
        theory_var v = mk_var(e);
        ctx.attach_th_var(e, this, v);
        ctx.mark_as_relevant(e);
        return true;
    }

    bool theory_jobscheduler::internalize_atom(app * atom, bool gate_ctx) {
        TRACE("csp", tout << mk_pp(atom, m) << "\n";);
        context & ctx = get_context();        
        SASSERT(u.is_model(atom));
        for (expr* arg : *atom) {
            if (!ctx.e_internalized(arg)) ctx.internalize(arg, false);
            internalize_cmd(arg);
        }
        add_done();
        bool_var bv = ctx.mk_bool_var(atom);
        ctx.set_var_theory(bv, get_id());
        return true;
    }

    // TBD: stronger parameter validation
    void theory_jobscheduler::internalize_cmd(expr* cmd) {
        symbol key, val;
        rational r;
        expr* job, *resource;
        unsigned j = 0, res = 0, cap = 0, loadpct = 100;
        time_t start = 0, end = std::numeric_limits<time_t>::max(), capacity = 0;
        properties ps;
        if (u.is_set_preemptable(cmd, job) && u.is_job(job, j)) {
            set_preemptable(j, true);            
        }
        else if (u.is_add_resource_available(cmd, resource, loadpct, start, end) && u.is_resource(resource, res)) {
            add_resource_available(res, loadpct, start, end, ps);
        }
        else if (u.is_add_job_resource(cmd, job, resource, loadpct, capacity, end) && u.is_job(job, j) && u.is_resource(resource, res)) {
            add_job_resource(j, res, loadpct, capacity, end, ps);
        }
        else {
            invalid_argument("command not recognized ", cmd);
        }
    }

    void theory_jobscheduler::invalid_argument(char const* msg, expr* arg) {
        std::ostringstream strm;
        strm << msg << mk_pp(arg, m);
        throw default_exception(strm.str());
    }

    void theory_jobscheduler::push_scope_eh() {
    }

    void theory_jobscheduler::pop_scope_eh(unsigned num_scopes) {
    }

    final_check_status theory_jobscheduler::final_check_eh() {
        TRACE("csp", tout << "\n";);
        bool blocked = false;
        for (unsigned r = 0; r < m_resources.size(); ++r) {
            if (constrain_resource_energy(r)) {
                blocked = true;
            }
        }
        for (unsigned j = 0; j < m_jobs.size(); ++j) {
            if (constrain_end_time_interval(j, resource(j))) {
                blocked = true;
            }
        }
        
        if (blocked) return FC_CONTINUE;
        return FC_DONE;
    }

    bool theory_jobscheduler::can_propagate() {
        return false;
    }

    literal theory_jobscheduler::mk_literal(expr * e) {
        expr_ref _e(e, m);
        context& ctx = get_context();
        if (!ctx.e_internalized(e)) {
            ctx.internalize(e, false);
        }
        ctx.mark_as_relevant(ctx.get_enode(e));
        return ctx.get_literal(e);
    }

    literal theory_jobscheduler::mk_ge(expr* e, time_t t) {
        return mk_literal(mk_ge_expr(e, t));
    }

    expr* theory_jobscheduler::mk_ge_expr(expr* e, time_t t) {
        return a.mk_ge(e, a.mk_int(rational(t, rational::ui64())));
    }

    literal theory_jobscheduler::mk_ge(enode* e, time_t t) {
        return mk_ge(e->get_owner(), t);
    }

    literal theory_jobscheduler::mk_le(expr* e, time_t t) {
        return mk_literal(mk_le_expr(e, t));
    }

    expr* theory_jobscheduler::mk_le_expr(expr* e, time_t t) {
        return a.mk_le(e, a.mk_int(rational(t, rational::ui64())));
    }

    literal theory_jobscheduler::mk_le(enode* l, enode* r) {
        context& ctx = get_context();
        expr_ref le(a.mk_le(l->get_owner(), r->get_owner()), m);
        ctx.get_rewriter()(le);
        return mk_literal(le);
    }

    literal theory_jobscheduler::mk_le(enode* e, time_t t) {
        return mk_le(e->get_owner(), t);
    }

    literal theory_jobscheduler::mk_eq_lit(expr * l, expr * r) {
        literal lit = mk_eq(l, r, false);
        get_context().mark_as_relevant(lit);
        return lit;
    }


    /**
     * iterator of job overlaps.
     */
    theory_jobscheduler::job_overlap::job_overlap(vector<job_time>& starts, vector<job_time>& ends):
        m_start(0), m_starts(starts), m_ends(ends), s_idx(0), e_idx(0) {
        job_time::compare cmp;
        std::sort(starts.begin(), starts.end(), cmp);
        std::sort(ends.begin(),   ends.end(),   cmp);        
    }

    bool theory_jobscheduler::job_overlap::next() {
        if (s_idx == m_starts.size()) {
            return false;
        }
        do {
            m_start = std::max(m_start, m_starts[s_idx].m_time);
                
            // add jobs that start before or at m_start
            while (s_idx < m_starts.size() && m_starts[s_idx].m_time <= m_start) {
                m_jobs.insert(m_starts[s_idx].m_job);
                ++s_idx;
            }
            // remove jobs that end before m_start.
            while (e_idx < m_ends.size() && m_ends[e_idx].m_time < m_start) {
                m_jobs.remove(m_ends[e_idx].m_job);
                ++e_idx;
            }   
        }
        // as long as set of jobs increments, add to m_start
        while (s_idx < m_starts.size() && e_idx < m_ends.size() && 
               m_starts[s_idx].m_time <= m_ends[e_idx].m_time);
            
        return true;
    }


    /**
     *  r = resource(j) & start(j) >= slb => end(j) >= ect(j, r, slb)
     */
    void theory_jobscheduler::propagate_end_time(unsigned j, unsigned r) {
        time_t slb = est(j);
        time_t clb = ect(j, r, slb);
        context& ctx = get_context();

        if (clb > end(j)) {
            job_info const& ji = m_jobs[j];
            literal start_ge_lo = mk_ge(ji.m_start, slb);
            if (ctx.get_assignment(start_ge_lo) != l_true) {
                return;
            }
            enode_pair eq(ji.m_job2resource, m_resources[r].m_resource);
            if (eq.first->get_root() != eq.second->get_root()) {
                return;
            }

            literal end_ge_lo = mk_ge(ji.m_end, clb);
            // Initialization ensures that satisfiable states have completion time below end.
            VERIFY(clb <= get_job_resource(j, r).m_end);
            region& r = ctx.get_region();
            ctx.assign(end_ge_lo, 
                       ctx.mk_justification(
                           ext_theory_propagation_justification(get_id(), r, 1, &start_ge_lo, 1, &eq, end_ge_lo, 0, nullptr)));
        }
    }

    /**
     *  For time interval [t0, t1] the end-time can be computed as a function 
     *  of start time based on reource load availability.
     * 
     *  r = resource(j) & t1 >= start(j) >= t0 => end(j) = start(j) + ect(j, r, t0) - t0
     */
    bool theory_jobscheduler::constrain_end_time_interval(unsigned j, unsigned r) {
        unsigned idx1 = 0, idx2 = 0;
        time_t s = start(j);
        TRACE("csp", tout << "job: " << j << " resource: " << r << " start: " << s <<  "\n";);
        if (!resource_available(r, s, idx1)) return false;
        vector<res_available>& available = m_resources[r].m_available;
        time_t e = ect(j, r, s);
        TRACE("csp", tout << "job: " << j << " resource: " << r << " ect: " << e <<  "\n";);
        if (!resource_available(r, e, idx2)) return false; // infeasible..
        time_t start1 = available[idx1].m_start;
        time_t end1   = available[idx1].m_end;
        unsigned cap1 = available[idx1].m_loadpct;
        time_t start2 = available[idx2].m_start;
        time_t end2   = available[idx2].m_end;
        unsigned cap2 = available[idx2].m_loadpct;
        // calculate minimal start1 <= t0 <= s,    such that ect(j, r, t0) >= start2 
        // calculate maximal s      <= t1 <= end1, such that ect(j, r, t1) <= end2
        time_t delta1 = (s - start1)*cap1;
        time_t delta2 = (e - start2)*cap2;
        time_t t0, t1;
        if (delta1 <= delta2) {
            t0 = start1;
        }
        else {
            // solve for t0:
            // (s - t0)*cap1 = (e - start2)*cap2;
            t0 = s - (delta2 / cap1); 
        }
        delta1 = (end1 - s)*cap1;
        delta2 = (end2 - e)*cap2;
        if (delta1 <= delta2) {
            t1 = end1;
        }
        else {
            // solve for t1:
            // (t1 - s)*cap1 = (end2 - e)*cap2
            t1 = s + (delta2 / cap1);
        }

        time_t delta = ect(j, r, t0) - t0;
        if (end(j) == start(j) + delta) {
            return false;
        }
        TRACE("csp", tout << "job: " << j << " resource " << r << " t0: " << t0 << " t1: " << t1 << " delta: " << delta << "\n";);
        literal_vector lits;
        enode* start_e = m_jobs[j].m_start;
        enode* end_e = m_jobs[j].m_end;
        lits.push_back(~mk_eq_lit(m_jobs[j].m_job2resource, m_resources[r].m_resource));
        lits.push_back(~mk_ge(start_e, t0));
        lits.push_back(~mk_le(start_e, t1));        
        expr_ref rhs(a.mk_add(start_e->get_owner(), a.mk_int(rational(delta, rational::ui64()))), m);
        lits.push_back(mk_eq_lit(end_e->get_owner(), rhs));
        context& ctx = get_context();
        ctx.mk_clause(lits.size(), lits.c_ptr(), nullptr, CLS_AUX_LEMMA, nullptr);
        return true;
    }

    void theory_jobscheduler::propagate_resource_energy(unsigned r) {
        
    }

    /**
     *  Ensure that job overlaps don't exceed available energy
     */ 
    bool theory_jobscheduler::constrain_resource_energy(unsigned r) {
        bool blocked = false;
        vector<job_time> starts, ends;
        res_info const& ri = m_resources[r];
        for (unsigned j : ri.m_jobs) {
            if (resource(j) == r) {
                starts.push_back(job_time(j, start(j)));
                ends.push_back(job_time(j, end(j)));
            }
        }
        job_overlap overlap(starts, ends);
        while (overlap.next()) {
            unsigned cap = 0;
            auto const& jobs = overlap.jobs();
            for (auto j : jobs) {
                cap += get_job_resource(j, r).m_loadpct;
                if (cap > 100) {
                    block_job_overlap(r, jobs, j);
                    blocked = true;
                    goto try_next_overlap;
                }
            }    
        try_next_overlap:
            ;
        }
        return blocked;
    }

    void theory_jobscheduler::block_job_overlap(unsigned r, uint_set const& jobs, unsigned last_job) {
        //
        // block the following case:
        // each job is assigned to r.
        // max { start(j) | j0..last_job } <= min { end(j) | j0..last_job }
        // joint capacity of jobs exceeds availability of resource.
        // 
        time_t max_start = 0;
        unsigned max_j = last_job;
        for (auto j : jobs) {
            if (max_start < start(j)) {
                max_start = start(j);
                max_j = j;
            }
            if (j == last_job) break;
        }
        literal_vector lits;
        for (auto j : jobs) {
            // create literals for:  
            // resource(j) == r
            // m_jobs[j].m_start <= m_jobs[max_j].m_start;
            // m_jobs[max_j].m_start <= m_jobs[j].m_end;
            lits.push_back(~mk_eq_lit(m_jobs[j].m_job2resource, m_resources[r].m_resource));
            if (j != max_j) {
                lits.push_back(~mk_le(m_jobs[j].m_start, m_jobs[max_j].m_start));
                lits.push_back(~mk_le(m_jobs[max_j].m_start, m_jobs[j].m_end));
            }
            if (j == last_job) break;
        }
        context& ctx = get_context();
        ctx.mk_clause(lits.size(), lits.c_ptr(), nullptr, CLS_AUX_LEMMA, nullptr);
    }

    void theory_jobscheduler::propagate() {
        return;
        for (unsigned j = 0; j < m_jobs.size(); ++j) {
            job_info const& ji = m_jobs[j];
            unsigned r = resource(j);
            propagate_end_time(j, r);
        }
        for (unsigned r = 0; r < m_resources.size(); ++r) {
            // TBD: check energy constraints on resources.
        }
    }
        
    theory_jobscheduler::theory_jobscheduler(ast_manager& m): 
        theory(m.get_family_id("csp")), m(m), u(m), a(m) {
    }

    std::ostream& theory_jobscheduler::display(std::ostream & out, job_resource const& jr) const {
        return out << "r:" << jr.m_resource_id << " cap:" << jr.m_capacity << " load:" << jr.m_loadpct << " end:" << jr.m_end << "\n";
    }

    std::ostream& theory_jobscheduler::display(std::ostream & out, job_info const& j) const {
        for (job_resource const& jr : j.m_resources) {
            display(out << "  ", jr);
        }
        return out;
    }

    std::ostream& theory_jobscheduler::display(std::ostream & out, res_available const& r) const {
        return out << "[" << r.m_start << ":" << r.m_end << "] @ " << r.m_loadpct << "%\n";
    }

    std::ostream& theory_jobscheduler::display(std::ostream & out, res_info const& r) const {
        for (res_available const& ra : r.m_available) {
            display(out << "   ", ra);
        }
        return out;
    }

    void theory_jobscheduler::display(std::ostream & out) const {
        out << "jobscheduler:\n";
        for (unsigned j = 0; j < m_jobs.size(); ++j) {
            display(out << "job " << j << ":\n", m_jobs[j]);
        }
        for (unsigned r = 0; r < m_resources.size(); ++r) {
            display(out << "resource " << r << ":\n", m_resources[r]);
        }
    }
        
    void theory_jobscheduler::collect_statistics(::statistics & st) const {

    }

    bool theory_jobscheduler::include_func_interp(func_decl* f) {
        return 
            f->get_decl_kind() == OP_JS_START ||
            f->get_decl_kind() == OP_JS_END ||
            f->get_decl_kind() == OP_JS_JOB2RESOURCE;
    }

    void theory_jobscheduler::init_model(model_generator & m) {

    }
    
    model_value_proc * theory_jobscheduler::mk_value(enode * n, model_generator & mg) {
        return alloc(expr_wrapper_proc, n->get_root()->get_owner());
    }

    bool theory_jobscheduler::get_value(enode * n, expr_ref & r) {
        std::cout << mk_pp(n->get_owner(), m) << "\n";
        return false;
    }

    theory * theory_jobscheduler::mk_fresh(context * new_ctx) { 
        return alloc(theory_jobscheduler, new_ctx->get_manager()); 
    }

    time_t theory_jobscheduler::get_lo(expr* e) {
        arith_value av(get_context());
        rational val;
        bool is_strict;
        if (av.get_lo(e, val, is_strict) && !is_strict && val.is_uint64()) {
            return val.get_uint64();
        }
        return 0;
    }

    time_t theory_jobscheduler::get_up(expr* e) {
        arith_value av(get_context());
        rational val;
        bool is_strict;
        if (av.get_up(e, val, is_strict) && !is_strict && val.is_uint64()) {
            return val.get_uint64();
        }
        return std::numeric_limits<time_t>::max();
    }

    time_t theory_jobscheduler::get_value(expr* e) {
        arith_value av(get_context());
        rational val;
        if (av.get_value(e, val) && val.is_uint64()) {
            return val.get_uint64();
        }
        return 0;
    }    

    time_t theory_jobscheduler::est(unsigned j) {
        return get_lo(m_jobs[j].m_start->get_owner());
    }

    time_t theory_jobscheduler::lst(unsigned j) {
        return get_up(m_jobs[j].m_start->get_owner());
    }

    time_t theory_jobscheduler::ect(unsigned j) {
        return get_lo(m_jobs[j].m_end->get_owner());
    }

    time_t theory_jobscheduler::lct(unsigned j) {
        return get_up(m_jobs[j].m_end->get_owner());
    }

    time_t theory_jobscheduler::start(unsigned j) {
        return get_value(m_jobs[j].m_start->get_owner());
    }

    time_t theory_jobscheduler::end(unsigned j) {
        return get_value(m_jobs[j].m_end->get_owner());
    }

    unsigned theory_jobscheduler::resource(unsigned j) {
        unsigned r;
        enode* next = m_jobs[j].m_job2resource, *n = next;
        do {
            if (u.is_resource(next->get_owner(), r)) {
                return r;
            }
            next = next->get_next();
        }
        while (next != n);
        return 0;
    }

    void theory_jobscheduler::set_preemptable(unsigned j, bool is_preemptable) {
        m_jobs.reserve(j + 1);
        m_jobs[j].m_is_preemptable = is_preemptable;        
    }

    void theory_jobscheduler::add_job_resource(unsigned j, unsigned r, unsigned loadpct, uint64_t cap, time_t end, properties const& ps) {
        SASSERT(get_context().at_base_level());
        SASSERT(1 <= loadpct && loadpct <= 100);
        SASSERT(0 < cap);
        m_jobs.reserve(j + 1);
        m_resources.reserve(r + 1);
        job_info& ji = m_jobs[j];
        if (ji.m_resource2index.contains(r)) {
            throw default_exception("resource already bound to job");
        }
        if (!ji.m_start) {
            context& ctx = get_context();
            app_ref job(u.mk_job(j), m);
            app_ref start(u.mk_start(j), m);
            app_ref end(u.mk_end(j), m);
            app_ref res(u.mk_job2resource(j), m);
            if (!ctx.e_internalized(job)) ctx.internalize(job, false);
            if (!ctx.e_internalized(start)) ctx.internalize(start, false);
            if (!ctx.e_internalized(end))   ctx.internalize(end, false);
            if (!ctx.e_internalized(res))   ctx.internalize(res, false);
            ji.m_job      = ctx.get_enode(job);
            ji.m_start    = ctx.get_enode(start);
            ji.m_end      = ctx.get_enode(end);
            ji.m_job2resource = ctx.get_enode(res);
        }
        ji.m_resource2index.insert(r, ji.m_resources.size());
        ji.m_resources.push_back(job_resource(r, cap, loadpct, end, ps));
        SASSERT(!m_resources[r].m_jobs.contains(j));
        m_resources[r].m_jobs.push_back(j);
    }

    void theory_jobscheduler::add_resource_available(unsigned r, unsigned max_loadpct, time_t start, time_t end, properties const& ps) {
        SASSERT(get_context().at_base_level());
        SASSERT(1 <= max_loadpct && max_loadpct <= 100);
        SASSERT(start <= end);
        m_resources.reserve(r + 1);
        res_info& ri = m_resources[r];
        if (!ri.m_resource) {
            context& ctx = get_context();
            app_ref res(u.mk_resource(r), m);
            if (!ctx.e_internalized(res)) ctx.internalize(res, false);            
            ri.m_resource = ctx.get_enode(res);
        }
        ri.m_available.push_back(res_available(max_loadpct, start, end, ps));
    }

    /*
     * Initialze the state based on the set of jobs and resources added.
     * For each job j, with possible resources r1, ..., r_n assert
     *   resource(j) = r_1 || resource(j) = r_2 || ... || resource(j) = r_n
     * For each job and resource r with deadline end(j,r) assert
     *   resource(j) = r => end(j) <= end(j,r)
     *
     * Ensure that the availability slots for each resource is sorted by time.
     */
    void theory_jobscheduler::add_done() {
        TRACE("csp", tout << "add-done begin\n";);
        context & ctx = get_context();
        
        // sort availability intervals
        for (unsigned r = 0; r < m_resources.size(); ++r) {
            res_info& ri = m_resources[r];
            vector<res_available>& available = ri.m_available;
            res_available::compare cmp;
            std::sort(available.begin(), available.end(), cmp);
        }

        expr_ref fml(m);
        literal lit, l1, l2, l3;

        for (unsigned j = 0; j < m_jobs.size(); ++j) {
            job_info const& ji = m_jobs[j];
            literal_vector disj;
            if (ji.m_resources.empty()) {
                throw default_exception("every job should be associated with at least one resource");
            }

            // start(j) <= end(j)            
            lit = mk_le(ji.m_start, ji.m_end);
            ctx.mk_th_axiom(get_id(), 1, &lit);

            time_t start_lb = std::numeric_limits<time_t>::max();
            time_t end_ub = 0;
            for (job_resource const& jr : ji.m_resources) {
                // resource(j) = r => end(j) <= end(j, r)
                // resource(j) = r => start(j) <= lst(j, r, end(j, r))
                unsigned r = jr.m_resource_id;
                res_info const& ri = m_resources[r];
                enode* j2r = m_jobs[j].m_job2resource;
                literal eq = mk_eq_lit(j2r, ri.m_resource);
                assert_last_end_time(j, r, jr, eq);
                assert_last_start_time(j, r, eq);
                disj.push_back(eq);
                start_lb = std::min(start_lb, ri.m_available[0].m_start);
                end_ub = std::max(end_ub, ri.m_available.back().m_end);

            }
            // resource(j) = r1 || ... || resource(j) = r_n
            ctx.mk_th_axiom(get_id(), disj.size(), disj.c_ptr());

            // start(j) >= start_lb
            lit = mk_ge(ji.m_start, start_lb);
            ctx.mk_th_axiom(get_id(), 1, &lit);

            // end(j) <= end_ub
            lit = mk_le(ji.m_end, end_ub);
            ctx.mk_th_axiom(get_id(), 1, &lit);
        }        
        for (unsigned r = 0; r < m_resources.size(); ++r) {
            res_info& ri = m_resources[r];
            vector<res_available>& available = ri.m_available;
            if (available.empty()) continue;
            enode* res = m_resources[r].m_resource;
            for (unsigned j : ri.m_jobs) {
                // resource(j) == r => start(j) >= available[0].m_start;
                enode* j2r = m_jobs[j].m_job2resource;
                assert_first_start_time(j, r, mk_eq_lit(j2r, res));
            }
            for (unsigned i = 0; i + 1 < available.size(); ++i) {
                if (available[i].m_end > available[i + 1].m_start) {
                    throw default_exception("availability intervals should be disjoint");
                }
                for (unsigned j : ri.m_jobs) {
                    // jobs start within an interval.
                    // resource(j) == r => start(j) <= available[i].m_end || start(j) >= available[i + 1].m_start;
                    enode* j2r = m_jobs[j].m_job2resource;
                    literal eq = mk_eq_lit(j2r, res);
                    assert_job_not_in_gap(j, r, i, eq);

                    // if job is not pre-emptable, start and end have to align within contiguous interval.
                    // resource(j) == r => end(j) <= available[i].m_end || start(j) >= available[i + 1].m_start
                    if (!m_jobs[j].m_is_preemptable && available[i].m_end + 1 < available[i+1].m_start) {
                        assert_job_non_preemptable(j, r, i, eq);
                    }
                }
            }
        }
        TRACE("csp", tout << "add-done end\n";);
    }

    void theory_jobscheduler::assert_last_end_time(unsigned j, unsigned r, job_resource const& jr, literal eq) {
        job_info const& ji = m_jobs[j];
        literal l2 = mk_le(ji.m_end, jr.m_end);
        get_context().mk_th_axiom(get_id(), ~eq, l2);
    }

    void theory_jobscheduler::assert_last_start_time(unsigned j, unsigned r, literal eq) {
        context& ctx = get_context();
        time_t t;
        if (lst(j, r, t)) {
            ctx.mk_th_axiom(get_id(), ~eq, mk_le(m_jobs[j].m_start, t));
        }
        else {
            eq.neg();
            ctx.mk_th_axiom(get_id(), 1, &eq);
        }
    }

    void theory_jobscheduler::assert_first_start_time(unsigned j, unsigned r, literal eq) {
        vector<res_available>& available = m_resources[r].m_available;
        literal l2 = mk_ge(m_jobs[j].m_start, available[0].m_start);
        get_context().mk_th_axiom(get_id(), ~eq, l2);
    }

    void theory_jobscheduler::assert_job_not_in_gap(unsigned j, unsigned r, unsigned idx, literal eq) {
        vector<res_available>& available = m_resources[r].m_available;        
        literal l2 = mk_ge(m_jobs[j].m_start, available[idx + 1].m_start);
        literal l3 = mk_le(m_jobs[j].m_start, available[idx].m_end);
        get_context().mk_th_axiom(get_id(), ~eq, l2, l3);        
    }

    void theory_jobscheduler::assert_job_non_preemptable(unsigned j, unsigned r, unsigned idx, literal eq) {
        vector<res_available>& available = m_resources[r].m_available;        
        literal l2 = mk_le(m_jobs[j].m_end, available[idx].m_end);
        literal l3 = mk_ge(m_jobs[j].m_start, available[idx + 1].m_start);
        get_context().mk_th_axiom(get_id(), ~eq, l2, l3);
    }

    

    /**
     * check that each job is run on some resource according to 
     * requested capacity.
     * 
     * Check that the sum of jobs run in each time instance
     * does not exceed capacity.
     */
    void theory_jobscheduler::validate_assignment() {
        vector<vector<job_time>> start_times, end_times;
        start_times.reserve(m_resources.size());
        end_times.reserve(m_resources.size());
        for (unsigned j = 0; j < m_jobs.size(); ++j) {
            unsigned r = resource(j);
            start_times[r].push_back(job_time(j, start(j))); 
            end_times[r].push_back(job_time(j, end(j))); 
            if (ect(j, r, start(j)) > end(j)) {
                throw default_exception("job not assigned full capacity");
            }
            unsigned idx;
            if (!resource_available(r, start(j), idx)) {
                throw default_exception("resource is not available at job start time");
            }
        }
        for (unsigned r = 0; r < m_resources.size(); ++r) {
            // order jobs running on r by start, end-time intervals
            // then consume ordered list to find jobs in scope.
            job_overlap overlap(start_times[r], end_times[r]);
            while (overlap.next()) {
                // check that sum of job loads does not exceed 100%
                unsigned cap = 0;
                for (auto j : overlap.jobs()) {
                    cap += get_job_resource(j, r).m_loadpct;
                }
                if (cap > 100) {
                    throw default_exception("capacity on resource exceeded");
                }
            }
        }
    }

    theory_jobscheduler::job_resource const& theory_jobscheduler::get_job_resource(unsigned j, unsigned r) const {
        job_info const& ji = m_jobs[j];
        return ji.m_resources[ji.m_resource2index[r]];
    }

    bool theory_jobscheduler::resource_available(unsigned r, time_t t, unsigned& idx) {
        vector<res_available>& available = m_resources[r].m_available;
        unsigned lo = 0, hi = available.size(), mid = hi / 2;
        while (lo < hi) {
            res_available const& ra = available[mid];
            if (ra.m_start <= t && t <= ra.m_end) {
                idx = mid;
                return true;
            }
            else if (ra.m_start > t && mid > 0) {
                hi = mid - 1;
                mid = lo + (mid - lo) / 2;
            }
            else if (ra.m_end < t) {
                lo = mid + 1;
                mid += (hi - mid) / 2;
            }
            else {
                break;
            }
        }
        return false;
    }


    /**
     * compute earliest completion time for job j on resource r starting at time start.
     */    
    time_t theory_jobscheduler::ect(unsigned j, unsigned r, time_t start) {
        job_resource const& jr = get_job_resource(j, r);
        vector<res_available>& available = m_resources[r].m_available;

        unsigned j_load_pct = jr.m_loadpct;
        time_t cap = jr.m_capacity;
        unsigned idx = 0;
        if (!resource_available(r, start, idx)) {
            TRACE("csp", tout << "resource is not available for " << j << " " << r << "\n";);
            return std::numeric_limits<time_t>::max();
        }
        SASSERT(cap > 0);
        
        for (; idx < available.size(); ++idx) {
            start             = std::max(start, available[idx].m_start);
            time_t end        = available[idx].m_end;
            unsigned load_pct = available[idx].m_loadpct;
            time_t delta = solve_for_capacity(load_pct, j_load_pct, start, end);
            TRACE("csp", tout << "delta: " << delta << " capacity: " << cap << " load " << load_pct << " jload: " << j_load_pct << " start: " << start << " end " << end << "\n";);
            if (delta > cap) {
                //
                // solve for end: 
                //     cap = load * (end - start + 1)
                // <=>
                //     cap / load = (end - start + 1)
                // <=>
                //     end = cap / load + start - 1
                //
                end = solve_for_end(load_pct, j_load_pct, start, cap);
                cap = 0;
            }
            else {
                cap -= delta;
            }
            if (cap == 0) {
                return end;
            }
        }        
        return std::numeric_limits<time_t>::max();        
    }

    time_t theory_jobscheduler::solve_for_end(unsigned load_pct, unsigned job_load_pct, time_t start, time_t cap) {
        SASSERT(load_pct > 0);
        SASSERT(job_load_pct > 0);
        //      cap = (load / job_load_pct) * (end - start + 1)
        // <=> 
        //      end - start + 1 = (cap * job_load_pct) / load 
        // <=>
        //      end = start - 1 + (cap * job_load_pct) / load
        // <=>
        //      end = (load * (start - 1) + cap * job_load_pct) / load
        unsigned load = std::min(load_pct, job_load_pct);
        return (load * (start - 1) + cap * job_load_pct) / load;
    }

    time_t theory_jobscheduler::solve_for_start(unsigned load_pct, unsigned job_load_pct, time_t end, time_t cap) {
        SASSERT(load_pct > 0);
        SASSERT(job_load_pct > 0);
        //      cap = (load / job_load_pct) * (end - start + 1)
        // <=> 
        //      end - start + 1 = (cap * job_load_pct) / load 
        // <=>
        //      start = end + 1 - (cap * job_load_pct) / load
        // <=>
        //      start = (load * (end + 1) - cap * job_load_pct) / load
        unsigned load = std::min(load_pct, job_load_pct);
        return (load * (end + 1) - cap * job_load_pct) / load;
    }

    time_t theory_jobscheduler::solve_for_capacity(unsigned load_pct, unsigned job_load_pct, time_t start, time_t end) {
        SASSERT(job_load_pct > 0);
        unsigned load = std::min(load_pct, job_load_pct);
        return (load * (end - start + 1)) / job_load_pct;
    }

    /**
     * Compute last start time for job on resource r.
     */
    bool theory_jobscheduler::lst(unsigned j, unsigned r, time_t& start) {
        start = 0;
        job_resource const& jr = get_job_resource(j, r);
        vector<res_available>& available = m_resources[r].m_available;
        unsigned j_load_pct = jr.m_loadpct;
        time_t cap = jr.m_capacity;
        for (unsigned idx = available.size(); idx-- > 0; ) {
            start    = available[idx].m_start;
            time_t end      = available[idx].m_end;
            unsigned load_pct = available[idx].m_loadpct;
            time_t delta = solve_for_capacity(load_pct, j_load_pct, start, end);
            if (delta > cap) {
                start = solve_for_start(load_pct, j_load_pct, end, cap);
                cap = 0;
            }
            else {
                cap -= delta;
            }
            if (cap == 0) {
                std::cout << "start " << start << "\n";
                return true;
            }
        }
        return false;
    }
    
};

